"""
Xplique
-------

The goal of Xplique is to provide a simple interface to the latest explanation
techniques
"""

__version__ = '1.2.1'

from . import attributions
from . import concepts
from . import features_visualizations
from . import commons
from . import plots

from .commons import Tasks
